from functools import wraps

from flask import Flask


def job(app: Flask, endpoint):
    def decorator(func):
        @wraps(func)
        def wrapped():
            func()
            return ""

        app.add_url_rule(
            f"/jobs/{endpoint}",
            func.__name__ + "_job",
            wrapped,
            strict_slashes=False,
            methods=["POST"],
        )

        return func

    return decorator



from typing import Union

from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/upload_grades")
def upload_grades(*, data: str):
    ...



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/paste_text")
def paste_text(*, data: str, name: str = None, is_private: bool = False):
    ...


@requires_master_secret
@service.route("/api/get_paste")
def get_paste(*, name: str) -> str:
    ...


def get_paste_url(name: str):
    return f"https://paste.cs61a.org/{name}"



import os
import urllib.parse

import flask
import requests
from flask import current_app, g, session, request, redirect, abort, jsonify
from flask_oauthlib.client import OAuth
from werkzeug import security
from urllib.parse import urlparse

from common.rpc.auth import get_endpoint
from common.rpc.secrets import get_secret
from common.url_for import get_host, url_for

AUTHORIZED_ROLES = ["staff", "instructor", "grader"]

REDIRECT_KEY = "REDIRECT_KEY"


def get_user():
    g.user_data = g.get("user_data") or current_app.remote.get("user")
    return g.user_data.data["data"]


def is_logged_in():
    return "access_token" in session


def is_staff(course):
    try:
        endpoint = get_endpoint(course=course)
        for participation in get_user()["participations"]:
            if participation["role"] not in AUTHORIZED_ROLES:
                continue
            if participation["course"]["offering"] != endpoint and endpoint is not None:
                continue
            return True
        return False
    except Exception as e:
        # fail safe!
        print(e)
        return False


def login():
    session[REDIRECT_KEY] = urlparse(request.url)._replace(netloc=get_host()).geturl()
    return redirect(url_for("login"))


def create_oauth_client(
    app: flask.Flask,
    consumer_key,
    secret_key=None,
    success_callback=None,
    return_response=None,
):
    oauth = OAuth(app)

    if os.getenv("ENV") == "prod":
        if secret_key is None:
            app.secret_key = get_secret(secret_name="OKPY_OAUTH_SECRET")
        else:
            app.secret_key = secret_key
    else:
        consumer_key = "local-dev-all"
        app.secret_key = "kmSPJYPzKJglOOOmr7q0irMfBVMRFXN"

    if not app.debug:
        app.config.update(
            SESSION_COOKIE_SECURE=True,
            SESSION_COOKIE_HTTPONLY=True,
            SESSION_COOKIE_SAMESITE="Lax",
        )

    remote = oauth.remote_app(
        "ok-server",  # Server Name
        consumer_key=consumer_key,
        consumer_secret=app.secret_key,
        request_token_params={"scope": "all", "state": lambda: security.gen_salt(10)},
        base_url="https://okpy.org/api/v3/",
        request_token_url=None,
        access_token_method="POST",
        access_token_url="https://okpy.org/oauth/token",
        authorize_url="https://okpy.org/oauth/authorize",
    )

    def check_req(uri, headers, body):
        """ Add access_token to the URL Request. """
        if "access_token" not in uri and session.get("access_token"):
            params = {"access_token": session.get("access_token")[0]}
            url_parts = list(urllib.parse.urlparse(uri))
            query = dict(urllib.parse.parse_qsl(url_parts[4]))
            query.update(params)

            url_parts[4] = urllib.parse.urlencode(query)
            uri = urllib.parse.urlunparse(url_parts)
        return uri, headers, body

    remote.pre_request = check_req

    @app.route("/oauth/login")
    def login():
        if app.debug:
            response = remote.authorize(callback=url_for("authorized", _external=True))
        else:
            response = remote.authorize(
                url_for("authorized", _external=True, _scheme="https")
            )
        return response

    @app.route("/oauth/authorized")
    def authorized():
        resp = remote.authorized_response()
        if resp is None:
            return "Access denied: error=%s" % (request.args["error"])
        if isinstance(resp, dict) and "access_token" in resp:
            session["access_token"] = (resp["access_token"], "")
            if return_response:
                return_response(resp)

        if success_callback:
            success_callback()

        target = session.get(REDIRECT_KEY)
        if target:
            session.pop(REDIRECT_KEY)
            return redirect(target)
        return redirect(url_for("index"))

    @app.route("/api/user", methods=["POST"])
    def client_method():
        if "access_token" not in session:
            abort(401)
        token = session["access_token"][0]
        r = requests.get("https://okpy.org/api/v3/user/?access_token={}".format(token))
        if not r.ok:
            abort(401)
        return jsonify(r.json())

    @remote.tokengetter
    def get_oauth_token():
        return session.get("access_token")

    app.remote = remote



import os
import pathlib
import sys
import time
import webbrowser
from base64 import b64encode
from os.path import relpath
from threading import Thread
from typing import Union

import click
import requests
from cachetools import LRUCache
from colorama import Fore, Style
from crcmod.crcmod import _usingExtension
from crcmod.predefined import mkPredefinedCrcFun
from tqdm import tqdm
from watchdog.events import (
    DirMovedEvent,
    FileMovedEvent,
    FileSystemEvent,
    FileSystemEventHandler,
)
from watchdog.observers import Observer

from common.cli_utils import pretty_print
from common.rpc.auth_utils import get_token
from common.rpc.sandbox import (
    get_server_hashes,
    initialize_sandbox,
    is_sandbox_initialized,
    run_make_command,
    update_file,
)
from common.shell_utils import sh

# TODO: Properly indicate if the extension is not installed


REPO = "berkeley-cs61a"

SYMLINK = "SYMLINK"
SUCCESS = "SUCCESS"

tracked_files = set()
remote_state = {}
recent_files = LRUCache(15)
do_build = True


def find_target():
    if not hasattr(find_target, "out"):
        remote = sh(
            "git",
            "config",
            "--get",
            "remote.origin.url",
            capture_output=True,
            quiet=True,
        ).decode("utf-8")
        if REPO not in remote:
            raise Exception(
                "You must run this command in the berkeley-cs61a repo directory"
            )
        find_target.out = (
            sh("git", "rev-parse", "--show-toplevel", capture_output=True, quiet=True)
            .decode("utf-8")
            .strip()
        )
    return find_target.out


def get_sandbox_url():
    username = requests.get(
        "https://okpy.org/api/v3/user/", params={"access_token": get_token()}
    ).json()["data"]["email"][: -len("@berkeley.edu")]
    return f"https://{username}.sb.cs61a.org"


@click.command()
def build():
    os.chdir(find_target())
    if not is_sandbox_initialized():
        print("Sandbox is not initialized.")
        if click.confirm(
            "You need to initialize your sandbox first. Continue?", default=True
        ):
            initialize_sandbox()
            for line in run_make_command(target="virtualenv"):
                print(line, end="")
            print()
        else:
            return
    print("Please wait until synchronization completes...")
    print("Scanning local directory...")
    full_synchronize_with_remote(get_server_hashes, show_progress=True)
    sandbox_url = get_sandbox_url()
    print(
        f"Synchronization completed! You can now begin developing. Preview your changes at {sandbox_url}"
    )
    webbrowser.open(sandbox_url)

    Thread(target=catchup_synchronizer_thread, daemon=True).start()
    Thread(target=catchup_full_synchronizer_thread, daemon=True).start()
    Thread(target=file_events_thread, daemon=True).start()

    # run this thread on the main thread to handle KeyboardInterrupts
    input_thread()


def input_thread():
    try:
        import readline
    except ImportError:
        # todo: use pyreadline on windows
        pass

    try:
        while True:
            print(Style.BRIGHT, end="")
            target = input("make> ")
            print(Style.RESET_ALL, end="")
            if not target.strip():
                continue
            try:
                for line in run_make_command(target=target):
                    print(line, end="")
            except KeyboardInterrupt:
                print(Fore.RED)
                pretty_print("🚫", "Build cancelled.")
            except Exception as e:
                print(Fore.RED)
                print(str(e))
                pretty_print("😿", "Build failed.")
            else:
                print(Fore.GREEN)
                pretty_print("🎉", "Build completed!")
    except KeyboardInterrupt:
        pretty_print("👋", "Interrupt signal received, have a nice day!")
        sys.exit(0)


def file_events_thread():
    while True:
        observer = Observer()
        try:
            observer.schedule(Handler(), find_target(), recursive=True)
            observer.start()
            for _ in range(15):
                time.sleep(1)
        finally:
            observer.stop()
            observer.join()


def catchup_synchronizer_thread():
    while True:
        recent_synchronization()
        time.sleep(1)


def catchup_full_synchronizer_thread():
    while True:
        full_synchronize_with_remote()
        time.sleep(15)


class Handler(FileSystemEventHandler):
    def on_any_event(self, event: FileSystemEvent):
        synchronize(event.src_path)

    def on_moved(self, event: Union[DirMovedEvent, FileMovedEvent]):
        synchronize(event.dest_path)


def synchronize(path: str):
    os.chdir(find_target())
    path = relpath(path)
    if isinstance(path, bytes):
        path = path.decode("ascii")
    if path not in tracked_files:
        # do not synchronize untracked files
        # if this is a file just created, it will be tracked on the next full_synchronization pass
        return
    recent_files[path] = None
    # path is a path to either a file or a symlink, NOT a directory
    if os.path.islink(path):
        update_file(path=path, symlink=os.readlink(path))
    elif os.path.isfile(path):
        with open(path, "rb") as f:
            update_file(
                path=path, encoded_file_contents=b64encode(f.read()).decode("ascii")
            )
    elif not os.path.exists(path):
        update_file(path=path, delete=True)
    remote_state[path] = get_hash(path)


def recent_synchronization():
    for path in set(recent_files.keys()):  # set() is needed to avoid concurrency issues
        if remote_state.get(path) != get_hash(path):
            synchronize(path)


def full_synchronize_with_remote(remote_state_getter=None, show_progress=False):
    global remote_state
    current_state = hash_all(show_progress=show_progress)
    if callable(remote_state_getter):
        # When we want to fetch it lazily
        if show_progress:
            print("Fetching server state...")
        remote_state = remote_state_getter()
    paths = set(current_state) | set(remote_state)
    to_update = []
    for path in paths:
        if current_state.get(path) != remote_state.get(path):
            to_update.append(path)

    for path in tqdm(to_update) if show_progress else to_update:
        synchronize(path)

    remote_state = current_state


def get_hash(path):
    assert _usingExtension, "You must use the crcmod C extension"
    if isinstance(path, bytes):
        path = path.decode("ascii")
    hash_func = mkPredefinedCrcFun("crc-32")
    if os.path.islink(path):
        return SYMLINK + os.readlink(path)
    try:
        with open(path, "rb") as f:
            return hash_func(f.read())
    except FileNotFoundError:
        return
    except IsADirectoryError:
        return


def hash_all(show_progress=False):
    global tracked_files
    files = (
        sh(
            "git", "ls-files", "--exclude-standard", capture_output=True, quiet=True
        ).splitlines()  # All tracked files
        + sh(
            "git",
            "ls-files",
            "-o",
            "--exclude-standard",
            capture_output=True,
            quiet=True,
        ).splitlines()  # Untracked but not ignored files
    )
    out = {}
    for file in tqdm(files) if show_progress else files:
        h = get_hash(file)
        if isinstance(file, bytes):
            file = file.decode("ascii")
        out[relpath(file)] = h
    tracked_files = set(out) | set(remote_state)
    return out



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/clear_queue")
def clear_queue(*, repo: str, pr_number: int):
    ...


@requires_master_secret
@service.route("/api/trigger_build")
def trigger_build_sync(*, pr_number: int, target_app: str = None):
    ...


@requires_master_secret
@service.route("/api/deploy_prod_app_sync")
def deploy_prod_app_sync(*, target_app: str):
    ...


@requires_master_secret
@service.route("/api/get_base_hostname")
def get_base_hostname(*, target_app: str) -> str:
    ...



import click, os

from sicp.build import build
from sicp.clone import clone
from sicp.venv import venv
from sicp.pr import pr
from sicp.auth import auth

from common.rpc.auth_utils import set_token_path


@click.group()
def cli():
    """
    This is an experimental general-purpose 61A task runner.
    """
    set_token_path(f"{os.path.expanduser('~')}/.sicp_token")


cli.add_command(clone)
cli.add_command(build)
cli.add_command(venv)
cli.add_command(pr)
cli.add_command(auth)

if __name__ == "__main__":
    cli()



from typing import Union

from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/export_attendance_rpc")
def rpc_export_attendance(*, full: bool):
    ...



import click

from common.shell_utils import sh
import os, shutil

ENV = "env"
REQ = "requirements.txt"


@click.command()
@click.argument("dir", default="./")
@click.argument("req", default="./")
@click.option("--reset", is_flag=True)
def venv(dir, req, reset):
    """Create a virtual environment in DIR.

    DIR is the location of the virtual env
    (minus the `env` folder itself). REQ is
    the location of the requirements file.
    Both of these arguments default to `./`.
    If you want to forcibly recreate an env,
    use the RESET option.
    """
    if not dir.endswith("/"):
        dir = dir + "/"
    if req.endswith("requirements.txt"):
        req = req[:-16]
    if not req.endswith("/"):
        req = req + "/"
    if os.path.exists(f"{dir}{ENV}"):
        if reset:
            shutil.rmtree(f"{dir}{ENV}")
        else:
            print("This environment already exists!")
            return
    sh("python3", "-m", "venv", f"{dir}{ENV}")
    sh(f"{dir}{ENV}/bin/pip3", "install", "-r", f"{req}{REQ}")



import click

from common.shell_utils import sh
import os, sys
from subprocess import CalledProcessError

HTTPS = "https://github.com/"
SSH = "git@github.com:"


@click.command()
@click.argument("repo")
@click.argument("dest", default="")
@click.option("--ssh", is_flag=True)
def clone(repo, dest, ssh):
    """Clone REPO into DEST.

    REPO is the name of the 61a repo to set up.
    Currently, "apps" and "cs61a" are supported.
    By default, DEST is set to the name of REPO.
    If you want to clone over SSH, use the SSH
    option.
    """
    if repo == "apps":
        run_apps_clone(dest if dest else "cs61a-apps", SSH if ssh else HTTPS)
    elif repo == "cs61a":
        run_61a_clone(dest if dest else "berkeley-cs61a", SSH if ssh else HTTPS)
    else:
        click.echo("No need to use sicp for that! Just git clone.", err=True)


def run_apps_clone(dir, protocol):
    print("========== Cloning cs61a-apps ==========")
    sh("git", "clone", f"{protocol}Cal-CS-61A-Staff/cs61a-apps", dir)

    print("========== Linking .githooks ===========")
    os.chdir(dir)
    sh("chmod", "+x", ".githooks/pre-commit")
    sh("git", "config", "core.hooksPath", ".githooks")

    print("====== Installing Black & Prettier =====")
    if "black" not in sh("pip3", "list", quiet=True).decode("utf-8"):
        sh("pip3", "install", "black")
    try:
        if "prettier" not in sh("npm", "list", "-g", quiet=True).decode("utf-8"):
            sh("sudo", "npm", "install", "-g", "prettier")
    except CalledProcessError:
        print(
            "Failed to install prettier globally as needed! "
            + "Make sure you have npm installed and have sudo privileges.",
            file=sys.stderr,
        )

    print("================ Done! =================")


def run_61a_clone(dir, protocol):
    print("======== Cloning berkeley-cs61a ========")
    sh("git", "clone", f"{protocol}Cal-CS-61A-Staff/berkeley-cs61a", dir)

    print("================ Done! =================")



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/index_piazza")
def index_piazza():
    ...


@requires_master_secret
@service.route("/api/clear_resources")
def clear_resources():
    ...


@requires_master_secret
@service.route("/api/upload_resources")
def upload_resources():
    ...



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/clear/piazza")
def clear_piazza():
    ...


@requires_master_secret
@service.route("/insert/piazza")
def insert_piazza(*, posts):
    ...


@requires_master_secret
@service.route("/clear/resources")
def clear_resources():
    ...


@requires_master_secret
@service.route("/insert/resources")
def insert_resources(*, resources):
    ...


@service.route("/query")
def query(*, piazza_params, resource_params):
    ...



from common.rpc.utils import create_service, requires_master_secret

service = create_service(__name__)


@requires_master_secret
@service.route("/api/add_domain")
def add_domain(*, course: str, domain: str):
    ...
